import abc
import hashlib

from mlpug.base import Base

from mlpug.utils.utils import get_value_at


class Callback(Base, metaclass=abc.ABCMeta):

    def __init__(self, name, base_logs_path="current", **kwargs):
        super(Callback, self).__init__(pybase_logger_name=name, **kwargs)

        self.name = name

        self.base_log_path = base_logs_path

        self.training_manager = None
        self.trainer = None
        self.model_components = None
        self.optimizers = None

    def set_training_manager(self, manager):
        self.training_manager = manager
        self.trainer = manager.get_trainer()
        self.model_components = self.trainer.get_model_components()
        self.optimizers = self.trainer.get_optimizers()

        return True

    def get_name(self):
        return self.name

    def get_hash(self):
        """
        Used to uniquely identify callback when loading and setting callback states
        :return:
        """
        hash_str = self._get_hash_string()

        return hashlib.md5(hash_str.encode('utf-8')).hexdigest()

    def get_state(self):
        """

        :return: state, success (True or False)
        """
        return None, True

    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """
        return True

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):
        """

        :param num_epochs:
        :param num_batches_per_epoch:
        :param start_epoch:
        :param start_batch:
        :param start_update_iter:

        :return: success (True or False)
        """
        return True

    def on_epoch_start(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        return True

    def on_batch_training_start(self, training_batch, logs):
        """

        :param training_batch:
        :param logs:

        :return: success (True or False)
        """
        return True

    def on_batch_training_failed(self, exception, logs):
        """

        :param exception:
        :param logs:

        :return: success (True or False)
        """
        return True

    def on_batch_training_completed(self, training_batch, logs):
        """

        :param training_batch:
        :param logs:

        :return: success (True or False)
        """
        return True

    def on_epoch_completed(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        return True

    def on_training_ended(self, stopped_early, stopped_on_error, interrupted, callback_calls_success):
        """

        :param stopped_early:
        :param stopped_on_error:
        :param interrupted:
        :param callback_calls_success:

        :return: success (True or False)
        """
        return True

    def _get_logs_base(self, logs):
        """
        Get the part of the log that needs to be accessed by the callback.
        Usually this is the "current" property, providing the logs for the current batch iteration.

        :param logs:
        :return:
        """

        if self.base_log_path is None:
            return logs
        else:
            return get_value_at(self.base_log_path, logs)

    def _get_hash_string(self):
        """
        A string, uniquely describing the callback
        :return:
        """
        return f"{self.__class__.__name__}={self.get_name()}={self._get_callback_properties_for_hash()}"

    def _get_callback_properties_for_hash(self):
        """
        This is used to create the unique callback hash.

        Returns a dict with properties describing the setup of the callback.
        It should at least contain the properties that influence the callback state.

        Property values should only be simple types, such as int, float, boolean and strings.
        Convert any object and function values (or similar) into a booleans (True = available, False = None)

        :return: dict
        """
        return {}




from enum import Enum

import math

from mlpug.trainers.callbacks.callback import Callback
from mlpug.mlpug_exceptions import CallbackInvalidException
from mlpug.utils import has_key, get_key_paths, SlidingWindow

from mlpug.evaluation import MetricEvaluatorBase

import basics.base_utils as _

from mlpug.utils import get_value_at, set_value_at


class MetricsLoggingMode(Enum):
    # TODO : Add BATCH_METRICS_ONLY Option
    BATCH_AND_WINDOW_AVERAGE_METRICS = "BATCH_AND_WINDOW_AVERAGE_METRICS"
    WINDOW_AVERAGE_METRICS = "WINDOW_AVERAGE_METRICS"
    WHOLE_DATASET_METRICS = "WHOLE_DATASET_METRICS"

    @staticmethod
    def will_log_window_average_metrics(mode):
        return mode in {MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS,
                        MetricsLoggingMode.WINDOW_AVERAGE_METRICS}

    def __str__(self):
        return str(self.value)


class MetricsLoggerBase(Callback):

    def __init__(self,
                 metric_evaluator,
                 dataset_name,
                 batch_level,
                 logging_mode,
                 dataset=None,
                 evaluate_settings=None,
                 batch_averaging_window=None,
                 log_condition_func=None,
                 sliding_window_factory=None,
                 name=None,
                 **kwargs):
        """

        Different modes of operation:

        Batch level: After every batch training iteration:
            Calculate metrics over a sliding window
                log metrics per batch and over window   batch_level = True, mode = BATCH_AND_WINDOW_AVERAGE_METRICS
                log metrics over window only            batch_level = True, mode = WINDOW_AVERAGE_METRICS
            Calculate metrics over whole data set       batch_level = True, mode = WHOLE_DATASET_METRICS

        Epoch level: After every training epoch:
            Calculate metrics over whole data set       batch_level = False, mode = WHOLE_DATASET_METRICS


        `log_condition_func` can be used to control if metrics are calculated

        :param dataset_name:
        :type dataset_name:

        :param batch_level: See general description above
        :type batch_level:

        :param logging_mode: See general description above
        :type logging_mode:

        :param dataset: Dataset, where the items are batches, to calculate the metrics with
        :type dataset:

        :param metric_evaluator: MetricEvaluator instance, or other object with the same public interface
        :type metric_evaluator:

        :param evaluate_settings: Default evaluation settings to use, when no evaluate_settings are available in the
                                  logs object provided to the callback methods
        :type evaluate_settings:

        :param batch_averaging_window: Window length in batches, over which the average metric must be calculated
        :type batch_averaging_window: Int

        :param sliding_window_factory: Optional. Function to create SlidingWindow instances

        :param name:
        :type name:
        """
        if name is None:
            name = self.__class__.__name__
            name = f"{name}[{dataset_name}]"

        super().__init__(name=name, **kwargs)

        self._dataset = dataset
        self._dataset_name = dataset_name

        self._metric_evaluator = metric_evaluator

        self._batch_level = batch_level

        self._logging_mode = logging_mode

        self._evaluate_settings = evaluate_settings

        self._batch_averaging_window = batch_averaging_window

        self._log_condition_func = log_condition_func or (lambda logs, dataset_batch: True)

        self._sliding_window_factory = SlidingWindow if not callable(sliding_window_factory) else sliding_window_factory

        self._name = name

        self._metric_windows = {}

        self._validate()

        if not self.instance_valid():
            raise CallbackInvalidException(self._name)

    def get_state(self):
        """

        :return: state, success (True or False)
        """
        return {
                    "metric_windows": {metric_path: window.get_state()
                                       for metric_path, window in self._metric_windows.items()}
               }, True

    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """

        if not self._check_state(state):
            self._log.error('Given state is invalid, unable to set state')
            return False

        self._metric_windows = {metric_path: self._sliding_window_factory(state=window_state)
                                for metric_path, window_state in state["metric_windows"].items()}

        return True

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        will_log_window_average_metrics = MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode)

        if not self.instance_valid():
            if will_log_window_average_metrics:
                self._log.error(f"{self} is not valid, unable to set up averaging window ... ")
            else:
                self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if will_log_window_average_metrics and \
                (self._batch_averaging_window is None or self._batch_averaging_window <= 0):
            self._log.error(f"A valid batch processing window is required for "
                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
            self._valid = False

            return False

        if will_log_window_average_metrics:
            self._init_metric_windows()

        return True

    def on_batch_training_completed(self, dataset_batch, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if not self._batch_level:
            return True

        if not self._log_condition_func(logs=logs, dataset_batch=dataset_batch):
            return True

        self._init_logs(logs)

        if self._logging_mode is MetricsLoggingMode.WHOLE_DATASET_METRICS:
            return self._calc_whole_dataset_metrics(logs, f"{self._dataset_name}.dataset")
        else:
            current = self._get_logs_base(logs)

            batch_metrics = {}
            if not self._calc_batch_metric_data_from(dataset_batch, batch_metrics, logs):
                return False

            base_path = f"{self._dataset_name}.batch"
            dataset_batch_logs = get_value_at(base_path, current)

            # Merge in new batch level results
            dataset_batch_logs = {**dataset_batch_logs, **batch_metrics}
            if self._logging_mode is MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS:
                set_value_at(base_path, current, dataset_batch_logs)

            metric_names = self._metric_evaluator.get_metric_names()
            metric_paths = get_key_paths(dataset_batch_logs,
                                         keys_to_consider=metric_names,
                                         keys_not_to_consider=["auxiliary_results"])

            self._update_metrics_windows_for(metric_paths, dataset_batch_logs, base_path=base_path)

            # gather all window data
            batch_metrics_lists = {p: s.window for p, s in self._metric_windows.items()}
            if not self._reduce(batch_metrics_lists, current[self._dataset_name]['window_average']):
                return False

            return True

    def on_epoch_completed(self, logs):
        if not self.instance_valid():
            self._log.error(f"{self} is not valid, skipping this hook ... ")
            return False

        if self._batch_level:
            return True

        if not self._log_condition_func(logs=logs, dataset_batch=None):
            return True

        self._init_logs(logs)

        return self._calc_whole_dataset_metrics(logs, f"{self._dataset_name}.dataset")

    def _init_logs(self, logs):
        current = self._get_logs_base(logs)
        if self._dataset_name not in current:
            current[self._dataset_name] = {}

        dataset_metrics = current[self._dataset_name]
        for metric_level in ['batch', 'window_average', 'dataset']:
            if metric_level not in dataset_metrics:
                dataset_metrics[metric_level] = {}

    def _get_current_evaluate_settings(self, logs):
        # Use custom settings if available, else use default settings

        evaluate_settings = get_value_at('evaluate_settings', logs, warn_on_failure=False)
        if evaluate_settings is None:
            evaluate_settings = self._evaluate_settings

        return evaluate_settings

    def _calc_whole_dataset_metrics(self, logs, log_path):

        current = self._get_logs_base(logs)
        metrics_log = get_value_at(log_path, current)

        evaluate_settings = self._get_current_evaluate_settings(logs)

        return self._metric_evaluator.calc_dataset_metrics_for(self._dataset,
                                                               metrics_log,
                                                               evaluate_settings=evaluate_settings,
                                                               dataset_name=self._dataset_name)

    def _init_metric_windows(self, reset=False):
        if reset:
            self._metric_windows = {}

        # Do nothing if metric windows are already available
        self._metric_windows = self._metric_windows or {}

        return True

    def _calc_batch_metric_data_from(self, batch, batch_metrics, logs):
        evaluate_settings = self._get_current_evaluate_settings(logs)

        model_output = None
        if self._dataset is None:
            current = self._get_logs_base(logs)
            loss = get_value_at(f"{self._dataset_name}.batch.loss", current)
            auxiliary_results = get_value_at(f"{self._dataset_name}.batch.auxiliary_results",
                                             current,
                                             warn_on_failure=False)

            model_output = {
                'loss': loss,
                'auxiliary_results': auxiliary_results
            }

        return self._metric_evaluator.calc_batch_metrics_for(batch,
                                                             batch_metrics,
                                                             evaluate_settings=evaluate_settings,
                                                             model_output=model_output)

    def _update_metrics_windows_for(self, metric_paths, batch_metrics, base_path):
        for metric_path in metric_paths:
            metric_value = get_value_at(metric_path, batch_metrics)

            full_metric_path = f"{base_path}.{metric_path}"
            sliding_window = self._metric_windows[metric_path] if metric_path in self._metric_windows else None
            if sliding_window is None:
                self._log.debug(f"Creating sliding window for {full_metric_path}")

                sliding_window = self._sliding_window_factory(length=self._batch_averaging_window, name=full_metric_path)
                self._metric_windows[metric_path] = sliding_window

            sliding_window.slide(metric_value)

    def _reduce(self, batch_metric_data_lists, reduced_metrics_log):

        return self._metric_evaluator.reduce(batch_metric_data_lists,
                                             reduced_metrics_log,
                                             dataset_name=self._dataset_name)

    def _get_callback_properties_for_hash(self):
        """
        This is used to create the unique callback hash.

        Returns a dict with properties describing the setup of the callback.
        It should at least contain the properties that influence the callback state.

        Property values should only be simple types, such as int, float, boolean and strings.
        Convert any object and function values (or similar) into a booleans
        (True = available, False = None, not available)

        :return: dict
        """
        props = super()._get_callback_properties_for_hash()
        return {
            **props,
            "metric_evaluator": self._metric_evaluator is not None,
            "dataset_name": self._dataset_name,
            "batch_level": self._batch_level,
            "logging_mode": self._logging_mode,
            "dataset": self._dataset is not None,
            "evaluate_settings": self._evaluate_settings is not None,
            "batch_averaging_window": self._batch_averaging_window,
            "log_condition_func": self._log_condition_func is not None,
            "sliding_window_factory": self._sliding_window_factory is not None
        }

    def _check_state(self, state):
        state_attributes = ['metric_windows']

        for attr in state_attributes:
            if not has_key(state, attr):
                self._log.error(f"Given state does not have a value for {attr}, state is invalid")
                return False

        return True

    def _validate(self):
        self._valid = True

        if not hasattr(self._dataset, '__iter__'):
            if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
                self._log.debug("Training batch evaluation results will be used to calculate and log metrics.")
            else:
                self._log.error(f"No valid dataset provided to calculated metrics on, "
                                f"the {self} will not function")
                self._valid = False

        if not MetricEvaluatorBase.is_valid(self._metric_evaluator):
            self._log.error(f"The given metric evaluator is not valid, the {self} will not function : "
                            f"{self._metric_evaluator}")
            self._valid = False

        if not type(self._batch_level) is bool:
            self._log.error(f"Batch level must be True or False, given value is: {self._batch_level}\n"
                            f"the {self} will not function")
            self._valid = False

        if self._logging_mode not in MetricsLoggingMode:
            self._log.error(f"Metric logging mode unknown ({self._logging_mode})\n"
                            f"the {self} will not function")
            self._valid = False

        if not self._batch_level and self._logging_mode is not MetricsLoggingMode.WHOLE_DATASET_METRICS:
            self._log.error(f"For epoch level metric logging, the metric logging can only be "
                            f"MetricsLoggingMode.WHOLE_DATASET_METRICS. value is :{self._logging_mode}\n"
                            f"the {self} will not function")
            self._valid = False

        if self._log_condition_func is not None and not callable(self._log_condition_func):
            self._log.error(f"The log condition function must be callable, "
                            f"metric logging mode ({self._logging_mode}, the {self} will not function")
            self._valid = False


class TrainingMetricsLogger(MetricsLoggerBase):
    """
    By default, gets already calculated loss and auxiliary results from logs

    TODO : needs testing
    TODO : what to do for batch_level=false? Allow usage of window average of metrics to be used for epoch level?
    """

    def __init__(self,
                 metric_evaluator,
                 batch_level=True,
                 logging_mode=MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS,
                 **kwargs):

        # print(f"mlpug.TrainingMetricsLogger(kwargs={kwargs})")

        super().__init__(
            metric_evaluator=metric_evaluator,
            dataset_name='training',
            batch_level=batch_level,
            logging_mode=logging_mode,
            **kwargs)

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
            if self._batch_averaging_window is None:
                self._log.debug(f"Set batch_averaging_window to number of batches per epoch : {num_batches_per_epoch}")
                self._batch_averaging_window = num_batches_per_epoch

            if self._batch_averaging_window == math.inf:
                self._log.error("The batch average window is infinite, unable to calculate window average. "
                                "Please set a finite batch_averaging_window during construction of this "
                                "TrainingMetricsLogger.")
                self._valid = False
                return False

        return super().on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)


class TestMetricsLogger(MetricsLoggerBase):
    """
    Child class must implement _evaluate_loss method

    TODO : We should implement a specialization for the most common case of logging the (window/dataset) loss average
    """

    def __init__(self,
                 dataset,
                 dataset_name,
                 metric_evaluator,
                 batch_level=True,
                 logging_mode=None,
                 **kwargs):

        if logging_mode is None:
            logging_mode = MetricsLoggingMode.BATCH_AND_WINDOW_AVERAGE_METRICS \
                if batch_level else MetricsLoggingMode.WHOLE_DATASET_METRICS

        super().__init__(dataset=dataset,
                         dataset_name=dataset_name,
                         metric_evaluator=metric_evaluator,
                         batch_level=batch_level,
                         logging_mode=logging_mode,
                         **kwargs)

        self._dataset_iterator = None

    def on_training_start(self,
                          num_epochs,
                          num_batches_per_epoch,
                          start_epoch,
                          start_batch,
                          start_update_iter):

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode):
            if self._batch_averaging_window is None:
                try:
                    self._batch_averaging_window = len(self._dataset)
                    if self._batch_averaging_window == math.inf:
                        self._log.error(f"The batch average window of the {self._dataset_name} data set is infinite, "
                                        f"unable to calculate window average. "
                                        f"Please set a finite batch_processing_window during construction of "
                                        f"this TestMetricsLogger.")
                        self._valid = False
                    else:
                        self._log.debug(f"Set batch averaging window to number of batches in "
                                        f"{self._dataset_name} data set : {self._batch_averaging_window}")
                except Exception as e:
                    _.log_exception(self._log, f"Unable to assess data set length to set the batch averaging window, "
                                               f"{self} will not function", e)
                    self._valid = False

        return super().on_training_start(num_epochs, num_batches_per_epoch, start_epoch, start_batch, start_update_iter)

    def on_epoch_start(self, logs):
        """

        :param logs:

        :return: success (True or False)
        """
        success = super().on_epoch_start(logs)

        if not success:
            self._log.error('A problem occurred, will not continue executing this hook')
            return success

        if MetricsLoggingMode.will_log_window_average_metrics(self._logging_mode) and \
                self._dataset_iterator is None:
            self._dataset_iterator = iter(self._dataset)

        return True

    def on_batch_training_completed(self, training_batch, logs):
        current = self._get_logs_base(logs)
        training_iter = current["global_iter"]

        if self._logging_mode is MetricsLoggingMode.WHOLE_DATASET_METRICS:
            return super().on_batch_training_completed(training_batch, logs)
        else:
            try:
                dataset_batch = next(self._dataset_iterator)
            except StopIteration:
                self._log.debug(f"[Iteration {training_iter}] Rewinding the {self._dataset_name} data set")
                self._dataset_iterator = iter(self._dataset)
                dataset_batch = next(self._dataset_iterator)

            return super().on_batch_training_completed(dataset_batch, logs)

    def _validate(self):
        if not hasattr(self._dataset, '__iter__'):
            self._log.error(f"The given dataset {str(self._dataset)} is not iterable, the {self} will not function")
            self._valid = False

        super()._validate()



from . import trainers
from .trainers import callbacks

from . import evaluation

import mlpug.mlpug_logging as logging

from mlpug.multi_processing import MultiProcessingManager
from mlpug.pytorch.multi_processing import MultiProcessingMixin, PyTorchDistributedContext

if not isinstance(MultiProcessingManager.get_context(), PyTorchDistributedContext):
    MultiProcessingManager.set_context(PyTorchDistributedContext())




import os
import sys
import time
import json

import abc

import pickle

from shutil import copyfile

from mlpug.utils import get_value_at
from mlpug.trainers.callbacks.callback import Callback

import basics.base_utils as _


class CheckpointManager(Callback, metaclass=abc.ABCMeta):

    def __init__(self,
                 model_hyper_parameters=None,
                 checkpoints_path="../trained-models/",
                 batch_level=True,
                 metric_to_monitor="validation.window_average.perplexity",
                 metric_opt_mode='min',
                 metric_monitor_period=None,  # batches or epochs
                 metric_checkpoint_threshold=None,
                 allow_checkpointing_at_start=False,
                 create_checkpoint_every=200,  # batches or epochs
                 archive_last_model_checkpoint_every=2000,  # batches or epochs
                 base_checkpoint_filename=time.strftime("%d-%m-%Y_%H-%M-%S"),
                 model_checkpoint_filename_ext="m-ckp",
                 training_checkpoint_filename_ext="t-ckp",
                 backup_before_override=True,
                 warn_on_model_quality_not_available=None,
                 name="CheckpointManager",
                 **kwargs):
        """

        :param model_hyper_parameters: Dict with hyper parameters of the model, this will be saved as part of
                                       the model checkpoint
        :param checkpoints_path:
        :param batch_level: True if monitoring happens on the scale of batches. If False, monitoring is done
                            on the epoch level
        :param metric_to_monitor: key path to metric value in the log object, e.g. `validation.batch.perplexity`
        :param metric_opt_mode: 'max', 'min'
        :param metric_monitor_period: If give, this is the period between checks for model quality improvement.
                                      This is in number of batches if `batch_level = True`, else it is a
                                      number of epochs.

                                      If not given, every time a value is available for the given metric_to_monitor
                                      the metric will be checked for improvement.

        :param metric_checkpoint_threshold: when given, the model quality must be below (or above, depending on
                                            metric_opt_mode) this threshold before a new best model checkpoint can be
                                            saved

        :param allow_checkpointing_at_start: Default False. When True: when a value for the metric_to_monitor is
                                             available at global batch iteration 0, checkpoints will be stored.

        :param create_checkpoint_every: period before saving next training/model checkpoint
                                        (will be overridden after each period)
                                        A 0 value disables this feature.

        :param archive_last_model_checkpoint_every: period before the last available model checkpoint is archived.
                                                    Period must be multiple of create_checkpoint_every period

                                                    A 0 value disables this feature.

        :param force_monitoring_on_epoch: When True, the given metric will also be monitored on every epoch
                                          in the case that monitoring level is batch level
        :param base_checkpoint_filename:
        :param model_checkpoint_filename_ext:
        :param training_checkpoint_filename_ext:
        :param backup_before_override: When True the training and model checkpoints are backed up before they are
                                       overridden by a new version. If backing up fails, no new version will be written
                                       to disk. This gives the user a chance to fix a problem, e.g. disk full, without
                                       interruption of the training process.

        :param warn_on_model_quality_not_available: Default is True when metric_monitor_period given else False.
                                                    If set to true and the metric_to_monitor is not available in
                                                    the logs, a warning will be logged.

                                                    This is useful for debugging purposes

        """
        super().__init__(name=name, **kwargs)

        self._model_hyper_parameters = model_hyper_parameters

        self._checkpoints_path = checkpoints_path
        self._model_checkpoint_filename_ext = model_checkpoint_filename_ext
        self._training_checkpoint_filename_ext = training_checkpoint_filename_ext
        self._base_filename = base_checkpoint_filename

        self._batch_level = batch_level

        self._metric_to_monitor = metric_to_monitor
        self._metric_opt_mode = metric_opt_mode

        self._metric_monitor_period = metric_monitor_period
        self._metric_checkpoint_threshold = metric_checkpoint_threshold
        self._allow_checkpointing_at_start = allow_checkpointing_at_start
        self._create_checkpoint_every = create_checkpoint_every
        self._archive_last_model_checkpoint_every = archive_last_model_checkpoint_every

        self._backup_before_override = backup_before_override

        if warn_on_model_quality_not_available is None:
            warn_on_model_quality_not_available = self._metric_monitor_period is not None

        self._warn_on_model_quality_not_available = warn_on_model_quality_not_available

        self._best_model_quality = float('Inf') if self._metric_opt_mode == 'min' else -float('Inf')
        self._best_model_iter = None

        self._check_settings()

        self._describe_setup()

    def get_state(self):
        return {
            "best_model_quality": self._best_model_quality,
            "best_model_iter": self._best_model_iter
        }, True

    def set_state(self, state):
        """

        :param state Dict with saved checkpoint state to continue tracking the best model during training
        :return:
        """
        success = True
        if not _.is_dict(state):
            self._log.debug("No initial checkpoint state given, starting with clean slate ...")
            return success

        self._log.debug("Using given initial checkpoint state: \n\n%s" % json.dumps(state, indent=4))

        try:
            self._best_model_quality = state['best_model_quality']
            self._best_model_iter = state['best_model_iter']
        except Exception as e:
            _.log_exception(self._log, "Unable to set checkpoint manager state", e)
            success = False

        return success

    def on_batch_training_completed(self, training_batch, logs):
        if not self._batch_level:
            return True

        return self._monitor('global_iter', logs)

    def on_epoch_completed(self, logs):
        create_latest_checkpoint = True
        if not self._batch_level:
            iter_name = 'epoch'
        else:
            # Monitor on epoch for batch level because checkpointing might be driven by
            # availability of metric_to_monitor value (which could also have been calculated specifically
            # at epoch ends.
            iter_name = 'global_iter'
            # ... but when monitoring on batch_level latest checkpoint might already be created
            create_latest_checkpoint = False

        self._log.debug("Epoch complete, checking if model quality improved ...")
        return self._monitor(iter_name, logs, create_latest_checkpoint=create_latest_checkpoint)

    def on_training_ended(self, stopped_early, stopped_on_error, interrupted, callback_calls_success):
        status = 'ended'
        if stopped_early:
            status = 'stopped early'

        if stopped_on_error:
            status = 'stopped on error'

        if interrupted:
            status = 'interrupted'

        sys.stdout.write("\n")
        sys.stdout.flush()
        self._log.info(f"Training {status}")
        success = True
        if self._create_checkpoint_every > 0:
            self._log.info(f"... storing latest model ...")
            latest_model_fname = self._create_model_checkpoint()
            success = (latest_model_fname is not None)

            if status == 'ended' or stopped_early or interrupted:
                self._log.info(f"... storing training checkpoint ...")
                training_checkpoint_fname = self._create_training_checkpoint()
                success &= (training_checkpoint_fname is not None)

        if abs(self._best_model_quality) != float('Inf') and self._best_model_iter is not None:
            iter_name = "global iteration" if self._batch_level else "epoch"

            self._log.info(f"Best model quality reached {self._metric_to_monitor}={self._best_model_quality} "
                           f"at {iter_name} {self._best_model_iter}")
        else:
            self._log.warn(f"No good enough model found, no best model checkpoint saved.")

        return success

    def training_checkpoint_file_name(self):
        b_fn = self._base_filename
        ext = self._training_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-training-checkpoint.{ext}')

    def current_model_file_name(self, training_iter):
        b_fn = self._base_filename
        ext = self._model_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-model-checkpoint-{training_iter}.{ext}')

    def latest_model_file_name(self):
        b_fn = self._base_filename
        ext = self._model_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-latest-model-checkpoint.{ext}')

    def best_model_file_name(self):
        b_fn = self._base_filename
        ext = self._model_checkpoint_filename_ext

        return os.path.join(self._checkpoints_path, f'{b_fn}-best-model-checkpoint.{ext}')

    def _describe_setup(self):
        self._log.info(f"Metric to monitor : {self._metric_to_monitor}")
        if self._metric_monitor_period is not None:
            self._log.info(f"Metric monitor period : {self._metric_monitor_period}")
        else:
            self._log.info(f"Metric monitor period : Metric checked whenever it is available")

            if self._allow_checkpointing_at_start:
                self._log.info(f"Will create checkpoints at first global iteration, if a metric value is available")

        time_scale = 'batches' if self._batch_level else 'epochs'
        self._log.info(f"Create last training & model checkpoints every "
                       f"{self._archive_last_model_checkpoint_every} {time_scale}")
        self._log.info(f"Archive last model checkpoint every "
                       f"{self._archive_last_model_checkpoint_every} {time_scale}")

    def _get_model_quality(self, current_logs):
        model_quality = get_value_at(self._metric_to_monitor,
                                     current_logs,
                                     warn_on_failure=self._warn_on_model_quality_not_available)

        if type(model_quality) is tuple:
            # use the first value as metric value, the other values are auxiliary results meant for other purposes
            model_quality = model_quality[0]

        return model_quality

    # TODO : it would maybe be better to split the monitoring and the checkpointing in to two separated methods
    def _monitor(self, iter_name, logs, create_latest_checkpoint=True):
        monitor_success = True
        data_saved = False

        current = self._get_logs_base(logs)
        training_iter = current[iter_name]
        if not self._allow_checkpointing_at_start and iter_name == "global_iter" and training_iter == 0:
            self._log.debug("Skipping checkpointing at first iteration")
            return monitor_success

        success, best_model_fname = self._monitor_for_best_checkpoint(iter_name, logs)
        monitor_success &= success
        data_saved |= best_model_fname is not None

        if create_latest_checkpoint:
            success, \
                latest_model_fname, \
                training_checkpoint_fname = self._monitor_for_latest_checkpoint(iter_name,
                                                                                logs,
                                                                                best_model_fname=best_model_fname)
            monitor_success &= success
            data_saved |= latest_model_fname is not None
            data_saved |= training_checkpoint_fname is not None

            success, \
                archived_model_fname = self._monitor_for_checkpoint_archiving(iter_name,
                                                                              logs,
                                                                              latest_model_fname=latest_model_fname)
            monitor_success &= success
            data_saved |= archived_model_fname is not None

        if data_saved:
            self._log.debug("Saving of data done.\n\n")

        return monitor_success

    def _monitor_for_best_checkpoint(self, iter_name, logs):
        current = self._get_logs_base(logs)

        training_iter = current[iter_name]

        pretty_iter_name = self.pretty_iter_name(iter_name)

        best_model_fname = None
        success = True

        check_model_quality = self._metric_monitor_period is None or \
                ((self._metric_monitor_period > 0) and (training_iter % self._metric_monitor_period == 0))

        if not check_model_quality:
            return success, best_model_fname

        model_quality = self._get_model_quality(current)

        model_quality_available = model_quality is not None
        model_quality_good_enough = True
        if model_quality_available and self._metric_checkpoint_threshold is not None:
            if self._metric_opt_mode == 'min':
                model_quality_good_enough = model_quality <= self._metric_checkpoint_threshold
            elif self._metric_opt_mode == 'max':
                model_quality_good_enough = model_quality >= self._metric_checkpoint_threshold

        if not model_quality_good_enough:
            self._log.debug("%s : %s : model quality not good enough : %s : %3e " % (pretty_iter_name,
                                                                                     training_iter,
                                                                                     self._metric_to_monitor,
                                                                                     model_quality))

        if model_quality_available and model_quality_good_enough:
            model_improved = ((self._metric_opt_mode == 'min') and (model_quality < self._best_model_quality)) or \
                             ((self._metric_opt_mode == 'max') and (model_quality > self._best_model_quality))

            if model_improved:
                sys.stdout.write("\n")
                sys.stdout.flush()

                self._log.info("%s : %s : Model improved : %s : %3e " % (pretty_iter_name,
                                                                         training_iter,
                                                                         self._metric_to_monitor,
                                                                         model_quality))

                if (self._best_model_iter is not None) and (training_iter < self._best_model_iter):
                    self._log.warn(f"Inconsistency: according to the current training {pretty_iter_name.lower()} "
                                   f"({training_iter}), current best model training {pretty_iter_name.lower()} "
                                   f"({self._best_model_iter}) is in the future. "
                                   f"Was the right training checkpoint loaded?")

                best_model_fname = self._save_current_model_as_best()
                if best_model_fname is not None:
                    self._best_model_quality = model_quality
                    self._best_model_iter = training_iter
                else:
                    self._log.error("Unable to save improved model checkpoint")
                    success = False

            self._update_logs(model_improved, logs, current)

        return success, best_model_fname

    def _monitor_for_latest_checkpoint(self, iter_name, logs, best_model_fname=None):
        current = self._get_logs_base(logs)
        training_iter = current[iter_name]

        success = True
        save_latest_model = (self._create_checkpoint_every > 0) and (training_iter % self._create_checkpoint_every == 0)

        if not save_latest_model:
            return success, None, None

        sys.stdout.write("\n")
        sys.stdout.flush()

        # Just copy best model if available
        latest_model_fname = self._create_model_checkpoint(file_to_copy=best_model_fname)
        success &= latest_model_fname is not None

        # Also save latest training checkpoint
        training_checkpoint_fname = self._create_training_checkpoint()
        training_checkpoint_success = (training_checkpoint_fname is not None)

        success &= training_checkpoint_success

        return success, latest_model_fname, training_checkpoint_fname

    def _monitor_for_checkpoint_archiving(self, iter_name, logs, latest_model_fname):
        current = self._get_logs_base(logs)
        training_iter = current[iter_name]

        archive_last_model = (self._archive_last_model_checkpoint_every > 0) and \
                             (training_iter % self._archive_last_model_checkpoint_every == 0)

        success = True
        if not archive_last_model:
            return success, None

        archived_model_fname = None
        if latest_model_fname is not None:
            sys.stdout.write("\n")
            sys.stdout.flush()

            model_fname = self.current_model_file_name(training_iter)
            copy_success = self._copy(latest_model_fname, model_fname)
            if copy_success:
                archived_model_fname = model_fname

            success &= copy_success
        else:
            self._log.error("Unable to create checkpoint for latest model, unable to archive latest model")
            success = False

        return success, archived_model_fname

    def _save_current_model_as_best(self):
        model_fn = self.best_model_file_name()

        if self._backup_before_override:
            try:
                self._backup_checkpoint(model_fn)
            except Exception as e:
                _.log_exception(self._log, f"A problem occurred backing up the last best model checkpoint, "
                                           f"will not override override model checkpoint with new one", e)
                return None

        return self._create_model_checkpoint(model_fn=model_fn)

    def _update_logs(self, model_improved, logs, current):
        if model_improved:
            logs["best"] = current.copy()

        current["is_best"] = model_improved

    def _create_model_checkpoint(self, model_fn=None, file_to_copy=None):
        if model_fn is None:
            model_fn = self.latest_model_file_name()

        if file_to_copy:
            if not self._copy(file_to_copy, model_fn):
                self._log.error(f"Unable to create model checkpoint based on file : {file_to_copy}")
                return None
        else:
            try:
                state, success = self._gather_model_checkpoint_data()
                if state is not None:
                    if not success:
                        self._log.warn("Gathering the model checkpoint data was not completely successful, "
                                       "will save available checkpoint data anyway ...")

                    self._log.debug(f"Saving model checkpoint : {model_fn}")
                    self._save_model_checkpoint(model_fn, state)
                else:
                    return None
            except Exception as e:
                _.log_exception(self._log, f"A problem occurred saving the latest model as checkpoint", e)
                return None

        return model_fn

    def _create_training_checkpoint(self):
        checkpoint_fname = self.training_checkpoint_file_name()

        if self._backup_before_override:
            try:
                if not self._backup_checkpoint(checkpoint_fname):
                    self._log.error("Unable to backup last training checkpoint, "
                                    "will not override override training checkpoint with new one")
                    return None
            except Exception as e:
                _.log_exception(self._log, f"A problem occurred backing up the last training checkpoint, "
                                           f"will not override override training checkpoint with new one", e)
                return None

        try:
            state, success = self._gather_training_checkpoint_data()
            if state is not None:
                if not success:
                    self._log.warn("Gathering the training checkpoint data was not completely successful, "
                                   "will save available checkpoint data anyway ...")

                self._log.debug(f"Saving training checkpoint : {checkpoint_fname}")
                self._save_training_checkpoint(checkpoint_fname, state)
            else:
                return None
        except Exception as e:
            _.log_exception(self._log, f"Unable to save training checkpoint", e)
            return None

        return checkpoint_fname

    def _gather_model_checkpoint_data(self):
        """

        :return: state, success
        """
        state, success = self.trainer.get_model_components_state()

        if state is not None:
            if not success:
                self._log.warn("Getting the model components state was not completely successful, "
                               "continuing anyway ...")

            if self._model_hyper_parameters is not None:
                state['hyper_parameters'] = self._model_hyper_parameters

            try:
                manager_state, manager_state_success = self.training_manager.get_state_for_model_checkpoint()
                state['manager_state'] = manager_state

                if not manager_state_success:
                    self._log.warn("Getting the manager state for the model checkpoint was not successful, "
                                   "will continue anyway ...")
                    success = False

            except Exception as e:
                _.log_exception(self._log, "Unable to add manager state to model checkpoint, "
                                           "continuing anyway ...", e)
                success = False
        else:
            # In any case when the state is None, gathering the model checkpoint data is not successful
            success = False

        return state, success

    def _gather_training_checkpoint_data(self):
        """

        :return: state, success
        """
        return self.training_manager.get_state()

    def _save_model_checkpoint(self, filename, state):
        with open(filename, 'wb') as f:
            pickle.dump(state, f)

    def _save_training_checkpoint(self, filename, state):
        with open(filename, 'wb') as f:
            pickle.dump(state, f)

    def _backup_checkpoint(self, filename):
        if os.path.isfile(filename):
            return self._copy(filename, f"{filename}.backup")
        else:
            return True

    def _copy(self, source_fname, dest_fname):
        try:
            self._log.debug("Copying model:\n[%s] ==> \n[%s]\n" % (source_fname, dest_fname))
            copyfile(source_fname, dest_fname)
        except Exception as e:
            _.log_exception(self._log, "Unable to copy [%s]" % source_fname, e)
            return False

        return True

    def _check_settings(self):
        if not os.path.exists(self._checkpoints_path):
            self._log.info(f"Creating checkpoint directory : {self._checkpoints_path}")
            os.makedirs(self._checkpoints_path)

        if (self._create_checkpoint_every < 0) and (self._archive_last_model_checkpoint_every > 0):
            self._log.error("archive_last_model_checkpoint_every can't be > 0 while _create_checkpoint_every < 0, "
                            "disabling archiving ... ")
            self._archive_last_model_checkpoint_every = -1

        if (self._create_checkpoint_every > 0) and \
           (self._archive_last_model_checkpoint_every > 0) and \
           (self._archive_last_model_checkpoint_every % self._create_checkpoint_every != 0):

            self._archive_last_model_checkpoint_every = 10 * self._create_checkpoint_every
            self._log.warn(f"archive_last_model_checkpoint_every must be "
                           f"exact multiple of _create_checkpoint_every, "
                           f"changing archive_last_model_checkpoint_every to "
                           f"[{self._archive_last_model_checkpoint_every}]")

    def _get_callback_properties_for_hash(self):
        """
        This is used to create the unique callback hash.

        Returns a dict with properties describing the setup of the callback.
        It should at least contain the properties that influence the callback state.

        Property values should only be simple types, such as int, float, boolean and strings.
        Convert any object and function values (or similar) into a booleans (True = available, False = None)

        :return: dict
        """
        props = super()._get_callback_properties_for_hash()
        return {
            **props,
            "batch_level": self._batch_level,
            "metric_to_monitor": self._metric_to_monitor,
            "metric_opt_mode": self._metric_opt_mode,
            "metric_monitor_period": self._metric_monitor_period,
            "metric_checkpoint_threshold": self._metric_checkpoint_threshold,
        }

    @staticmethod
    def pretty_iter_name(iter_name):
        if iter_name == "epoch":
            return "Epoch"
        elif iter_name == "global_iter":
            return "Global iter"
        else:
            return iter_name



import abc

from mlpug.utils import get_value_at, convert_to_dict
from mlpug.trainers.callbacks.callback import Callback

import basics.base_utils as _


class LRSchedulerWrapperBase(Callback, metaclass=abc.ABCMeta):

    def __init__(self,
                 schedulers,
                 batch_level=True,
                 metric_to_monitor=None,
                 name="LRSchedulerWrapper",
                 **kwargs):
        """

        Calls given schedulers, per batch or per epoch, optionally using the current model quality
        as given by `metric_to_monitor`. The updated learning rate is added to the logs object.

        :param schedulers: A single scheduler instance or a dict or list of schedulers
        :param batch_level: True if the LR schedulers should be updated after every batch, else after every epoch
        :param metric_to_monitor: key path to metric value in the log object,
                                  e.g. `validation.window_average.perplexity`, used by the schedulers. If None, it is
                                  assumed that the schedulers don't monitor any metric.
        """
        super().__init__(name=name, **kwargs)

        schedulers = convert_to_dict("scheduler", schedulers)

        self._schedulers = schedulers
        self._batch_level = batch_level

        self._metric_to_monitor = metric_to_monitor

    @abc.abstractmethod
    def get_state(self):
        """

        :return: state, success (True or False)
        """
        self._log.error("This method is not implemented, implement it in your child class implementation")
        return None, False

    @abc.abstractmethod
    def set_state(self, state):
        """

        :param state:
        :return: success (True or False)
        """
        self._log.error("This method is not implemented, implement it in your child class implementation")
        return None, False

    def on_epoch_start(self, logs):
        if self._batch_level:
            return True

        self._init_logs(logs)

        return self._update_logs(logs)

    def on_batch_training_start(self, training_batch, logs):
        if not self._batch_level:
            return True

        self._init_logs(logs)

        return self._update_logs(logs)

    def on_batch_training_completed(self, training_batch, logs):
        if not self._batch_level:
            return True

        self._init_logs(logs)

        return self._update_lr(logs, 'global_iter')

    def on_epoch_completed(self, logs):
        if self._batch_level:
            return True

        self._init_logs(logs)

        return self._update_lr(logs, 'epoch')

    def _get_schedule_level(self):
        return 'batch' if self._batch_level else 'epoch'

    def _init_logs(self, logs):
        schedule_level = self._get_schedule_level()

        current = self._get_logs_base(logs)
        ctp = current['training_params'][schedule_level]

        if 'lr' not in ctp:
            ctp['lr'] = {}

    def _update_lr(self, logs, iter_name):
        current = self._get_logs_base(logs)

        model_quality = self._get_model_quality(current)
        training_iter = current[iter_name]

        success = False
        try:
            success = self._exec_schedulers(training_iter, model_quality)
            if not success:
                self._log.error("Updating of the learning rate(s) failed.")
        except Exception as e:
            _.log_exception(self._log, "An unexpected error occurred, "
                                       "execution of the learning rate scheduler(s) failed", e)

        return self._update_logs(logs) and success

    def _update_logs(self, logs):
        current = self._get_logs_base(logs)

        schedule_level = self._get_schedule_level()
        ctp = current['training_params'][schedule_level]

        try:
            current_lr = self._get_current_lr()

            lr = get_value_at('lr', ctp, warn_on_failure=False) or {}

            ctp['lr'] = {**lr, **current_lr}

            return True
        except Exception as e:
            _.log_exception(self._log, "An unexpected error occurred, "
                                       "unable to add current learning rate values to the logs object", e)
            return False

    def _get_model_quality(self, current_logs):
        if not self._metric_to_monitor:
            return None

        model_quality = get_value_at(self._metric_to_monitor, current_logs)

        if type(model_quality) is tuple:
            # use the first value as metric value, the other values are auxiliary results meant for other purposes
            model_quality = model_quality[0]

        return model_quality

    @abc.abstractmethod
    def _exec_schedulers(self, training_iter, model_quality=None):
        """

        :param training_iter
        :param model_quality:
        :return: On success : True
                 On failure : False
        """
        self._log.error("This method is not implemented, implement it in your child class implementation")
        return False

    @abc.abstractmethod
    def _get_current_lr(self):
        """
        :return: dict with learning rate, per optimizer
        """
        self._log.error("This method is not implemented, implement it in your child class implementation")
        return None

    def _get_callback_properties_for_hash(self):
        """
        This is used to create the unique callback hash.

        Returns a dict with properties describing the setup of the callback.
        It should at least contain the properties that influence the callback state.

        Property values should only be simple types, such as int, float, boolean and strings.
        Convert any object and function values (or similar) into a booleans
        (True = available, False = None, not available)

        :return: dict
        """
        props = super()._get_callback_properties_for_hash()
        return {
            **props,
            "batch_level": self._batch_level,
            "metric_to_monitor": self._metric_to_monitor
        }



